/*******************************************************************************
* FILE NAME: auto_drive_routines.c
*
* DESCRIPTION:
*   Autonomous drive functions
*******************************************************************************/

#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "stdio.h"

#include "auto_routines.h"
#include "user_routines.h"
#include "clock_routines.h"
#include "gyro.h"
#include "utilities.h"



/*******************************************************************************
* FUNCTION NAME: Auto_Hold_Position
* PURPOSE:       Apply motor power to hold position if necessary.
*                This is in force automatically whenever the robot is supposed 
*                to be at rest and acts as a brake.
*******************************************************************************/
void Auto_Hold_Position()
{
  if (holdflag)
  {
     Hold_Position((int)FALSE, (int)0, (int)0); // Hold this position even if pushed
  }
  else
  { // Initialize
    Hold_Position((int)TRUE, (int)0, (int)0);

    /* Disable interrupts while we zero the shaft encoders */
    INTCONbits.GIEL = 0;
    Right_Encoder = 0;
    Left_Encoder = 0;
    INTCONbits.GIEL = 1;
    holdflag = TRUE;
  }
}


/*******************************************************************************
* FUNCTION NAME: Drive_Forward_Time
* PURPOSE:       No sensor substitute for driving straight for a given time at a given speed.
*                Each function determines completion or failure.
* ARGUMENTS:     input: time in tenths of seconds, speed 0-127
*******************************************************************************/
void Drive_Forward_Time(int time, int speed)
{
  static unsigned long duration; //in tenths of seconds

//  printf("DFWD\r");

  if (lastDriveCmd != DFWD)
  {
    printf("DFWD 1\r");
    lastDriveCmd = DFWD;
    duration = Clockds + time; // Current time + duration = cmd completed
  }

  if (duration > Clockds)
  { // Keep going
    Drive_Status = INCOMPLETE;

    /* Might want to balance the motors here */
    WHL_L1 = -speed;
    WHL_R1 = WHL_L1; 
  }
  else
  { // All done (move on to next drive command)
    Drive_Status = COMPLETE;
    lastDriveCmd = ENDCMD;  // In case there are two of the same type cmd in a row
    WHL_L1 = NEUTRAL;
    WHL_R1 = NEUTRAL;
  }
}

/*******************************************************************************
* FUNCTION NAME: Drive_Backward_Time
* PURPOSE:       No sensor substitute for driving straight backwards
* ARGUMENTS:     input: time in tenths of seconds, speed 0-127
*******************************************************************************/
void Drive_Backward_Time(int time, int speed)
{
  static unsigned long duration; //in tenths of seconds

  if (lastDriveCmd != DBCK)
  {
    lastDriveCmd = DBCK;
    duration = Clockds + time; // Current time + duration = cmd completed
  }

  if (duration > Clockds)
  { // Keep going
    Drive_Status = INCOMPLETE;

    /* Might want to balance the motors here */
    WHL_L1 = speed;
    WHL_R1 = WHL_L1;
  }
  else
  { // All done (move on to next drive command)
    Drive_Status = COMPLETE;
    lastDriveCmd = ENDCMD;  // In case there are two of the same type cmd in a row
    WHL_L1 = NEUTRAL;
    WHL_R1 = NEUTRAL;
  }
}



/*******************************************************************************
* FUNCTION NAME: Drive_Spin_In_Place_Time
* PURPOSE:       No sensor substitute for a gyro turn
* ARGUMENTS:     input: time in tenths of seconds, speed 0-127, direction clockwise or CCW
*******************************************************************************/
void Drive_Spin_In_Place_Time(int time, int speed, int direction)
{
  static unsigned long duration; //in tenths of seconds

  if (lastDriveCmd != DSPIN)
  {
    lastDriveCmd = DSPIN;
    Drive_Status = INCOMPLETE;
    duration = Clockds + time; // Current time + duration = cmd completed
  }

  if (duration > Clockds)
  { // Keep going
    Drive_Status = INCOMPLETE;

    if (direction == CW) /* Spin Clockwise */
    {
      WHL_L1 = -speed;
      WHL_R1 = speed;

    }else /* Spin the opposite direction - Counter Clockwise */
    {
      WHL_L1 = speed;
      WHL_R1 = -speed;
    }
  }
  else
  { // All done (move on to next drive command)
    Drive_Status = COMPLETE;
    lastDriveCmd = ENDCMD;  // In case there are two of the same type cmd in a row
    WHL_L1 = NEUTRAL;
    WHL_R1 = NEUTRAL;
  }
}



/*******************************************************************************
* FUNCTION NAME: Drive_Stop
* PURPOSE:       No sensor substitute for driving straight for a given time at a given speed.
*                Each function determines completion or failure.
* ARGUMENTS:     input: time in tenths of seconds, speed 0-127
*******************************************************************************/
void Drive_Stop()
{
  WHL_L1 = NEUTRAL;
  WHL_R1 = NEUTRAL;

  // All done (move on to next drive command)
  Drive_Status = COMPLETE;
  lastDriveCmd = ENDCMD;  // In case there are two of the same type cmd in a row
}



/************************************************************   *******************
* FUNCTION NAME: Gyro_Turn
* PURPOSE:       Turn the robot to the specified heading.
* ARGUMENTS:     input: Target Heading                 
*******************************************************************************/
void Gyro_Turn(signed int TargetHeading)
{
  static signed long rawTargetHeading = 0;
//  printf("Gyro Turn Starts Here\r");
  if (lastDriveCmd != GTRN)
  {
    lastDriveCmd = GTRN;
    rawTargetHeading = GYRO_DEG_RAW(TargetHeading);
//    printf("Gyro_Turn1: rawTargetHeading=%ld, RawHeading=%ld, WHL_R1=%d\r", rawTargetHeading, RawHeading, WHL_R1);
  }
  if (abs(RawHeading - rawTargetHeading) > GTURN_DEADBAND)
  { // Keep going
    Drive_Status = INCOMPLETE;
  
    /* Might want to balance the motors here */
    WHL_R1 = GyroTurnRaw(rawTargetHeading);
    WHL_L1 = WHL_R1;
//    printf("Gyro_Turn2: rawTargetHeading=%ld, RawHeading=%ld, WHL_R1=%d\r", rawTargetHeading, RawHeading, WHL_R1);
  }
  else
  { // All done (move on to next drive command)
    Drive_Status = COMPLETE;
    lastDriveCmd = ENDCMD;  // In case there are two of the same type cmd in a row
    WHL_L1 = NEUTRAL;
    WHL_R1 = NEUTRAL;
  }
}



/*******************************************************************************
* FUNCTION NAME: Gyro_Turn_Sloppy
* PURPOSE:       Turn the robot to the specified heading.
* ARGUMENTS:     input: Target Heading                 
*******************************************************************************/
void Gyro_Turn_Sloppy(signed int TargetHeading)
{
  static signed long rawTargetHeading = 0;
  static char positive_direction;
  long error;

  printf("Gyro Turn Sloppy Starts Here\r");
  if (lastDriveCmd != GTRNS)
  {
    lastDriveCmd = GTRNS;
    rawTargetHeading = GYRO_DEG_RAW(TargetHeading);
    error = RawHeading - rawTargetHeading;
    if (error > 0)
        positive_direction = TRUE;
    else
        positive_direction = FALSE;
    
    printf("Gyro_Turn_Sloppy1: rawTargetHeading=%ld, RawHeading=%ld, WHL_R1=%d\r", rawTargetHeading, RawHeading, WHL_R1);
  }
  error = RawHeading - rawTargetHeading;

  if (((error > GTURN_DEADBAND) && (positive_direction==TRUE)) ||
      ((error < -GTURN_DEADBAND) && (positive_direction==FALSE)))
  { // Keep going
    Drive_Status = INCOMPLETE;
  
    /* Might want to balance the motors here */
    WHL_R1 = GyroTurnRaw(rawTargetHeading);
    WHL_L1 = -WHL_R1;
    printf("Gyro_Turn_Sloppy2: rawTargetHeading=%ld, RawHeading=%ld, WHL_R1=%d\r", rawTargetHeading, RawHeading, WHL_R1);
  }
  else
  { // All done (move on to next drive command)
    Drive_Status = COMPLETE;
    lastDriveCmd = ENDCMD;  // In case there are two of the same type cmd in a row
    WHL_L1 = NEUTRAL;
    WHL_R1 = NEUTRAL;
  }
}

#if 0  
/*******************************************************************************
* FUNCTION NAME: Drive_Forward_Time_Gyro
* PURPOSE:       Drive straight for a given time at a given speed.
*                using the gyro to remain straight.
*                Each function determines completion or failure.
* ARGUMENTS:     input: time in tenths of seconds, speed 0-127
*******************************************************************************/
void Drive_Forward_Time_Gyro(int time, int speed)
{
  static unsigned long duration; //in tenths of seconds
 printf("DFGWD\r");
  if (lastDriveCmd != DFGWD)
  {
    lastDriveCmd = DFGWD;
    duration = Clockds + time; // Current time + duration = cmd completed
  }

  if (duration > Clockds)
  { // Keep going
    Drive_Status = INCOMPLETE;

    /* Might want to balance the motors here */
    WHL_L1 = GyroDrive(speed);
    WHL_R1 = WHL_L1; 
  }
  else
  { // All done (move on to next drive command)
    Drive_Status = COMPLETE;
    lastDriveCmd = ENDCMD;  // In case there are two of the same type cmd in a row
    WHL_L1 = NEUTRAL;
    WHL_R1 = NEUTRAL;
  }
}
#endif


/*******************************************************************************
* FUNCTION NAME: Drive_Forward_Dist
* PURPOSE:       Drive straight for a given distance at a given speed.
*                Each function determines completion or failure.
* ARGUMENTS:     input: time in tenths of seconds, speed 0-127
*******************************************************************************/
#define DIST_KP  30/10

void Drive_Forward_Dist(int dist, int speed)
{
  static long distance; //in hundredths of inches
  long error;

  if (lastDriveCmd != DISTFWD)
  {
    printf("DISTFWD 1\r");
    lastDriveCmd = DISTFWD;
    distance = dist * (long)TICKS_PER_REV / (long)ENCODER_INCHES; // Convert from inches to encoder ticks

    /* Disable interrupts while we zero the shaft encoders */
    INTCONbits.GIEL = 0;
    Left_Encoder = Right_Encoder = 0;
    INTCONbits.GIEL = 1;
  }

  error = distance - (Right_Encoder + Left_Encoder)/2;
  if (error > 0)
  { // Keep going
    Drive_Status = INCOMPLETE;

    /* Might want to balance the motors here */
//    WHL_L1 = -speed;
    WHL_L1 = -error * DIST_KP;
    if (WHL_L1 > speed) WHL_L1 = speed;
    WHL_R1 = WHL_L1; 
  }
  else
  { // All done (move on to next drive command)
    Drive_Status = COMPLETE;
    lastDriveCmd = ENDCMD;  // In case there are two of the same type cmd in a row
    WHL_L1 = NEUTRAL;
    WHL_R1 = NEUTRAL;
  }
//  printf("error=%ld, distance=%ld, RE=%ld, LE=%ld\r", error, distance, Right_Encoder, Left_Encoder);
}
