/*******************************************************************************
* FILE NAME: targeting.c
*
* DESCRIPTION:
*  This file contains targeting software routines.
*******************************************************************************/

#include <stdio.h>

#include "ifi_utilities.h"
#include "ifi_aliases.h"
#include "ifi_default.h"
#include "utilities.h"
#include "user_routines.h"
#include "velocity_control.h"
#include "tracking.h"
#include "camera.h"
#include "clock_routines.h"
#define VARIABLES_DECLARE
	#include "targeting.h"
#undef VARIABLES_DECLARE

extern unsigned char camera_initialized;
extern unsigned int  camera_t_packets;

#define LEFT  1
#define RIGHT 0

char Last_Seen=RIGHT; // In what direction was the light last detected

char Last_Sighting=0; // Last target sighting: 0 = none, 1=left, 2=right

#define LAUNCH_RELAY relay1_fwd
#define TARGETING_DEADBAND 8 // When we can fire - 1/2 pixels equivalent to 33" goal 16pixels@25' & 36@5'

//#define BALL_SWITCH rc_dig_in01
//#define REED_SWITCH rc_dig_in02
//#define PRESSURE_GAUGE Get_Analog_Value(rc_ana_in03)
//#define PRESSURE_REQ 358
#define TARGET_HIT_THRESH 1  // Must have this many consecutive on_targets to be considered locked on target

char GimbalTrackPan(int *pan_motor)
{
    static unsigned int old_camera_t_packets;  // Local check
    static unsigned char packet_err_cnt = 0;
    signed int error;
    static char target_lock;  // Are we sure we're locked onto the target and not just passing it
	static char init=0;
	static char on_target=0;  // Applies to current or most recent camera T-packet report
    static unsigned char target_hits=0;
    static int old_pan_motor=0;
    static char search_init=TRUE;


//	printf("Welcome to GimbalTrackPan\r");
  *pan_motor = old_pan_motor;
  if(!init)
  {
		error = 0;
		on_target = 0;	
		init++;
  }

  if (!camera_initialized)  // If camera isn't on yet or is restarting, then tracking is not possible
  {
    on_target = FALSE;
    if ((Clockms%8) == 0)  // Flash LEDs until camera is initialized-Clockms is updated in 4ms increments
    {
        Relay2_red   =  Relay2_green;
        Relay2_green = !Relay2_green;
        rc_dig_out03 = !rc_dig_out03;
    }
  }
  else
  {
    rc_dig_out03 = ON;
    if(camera_t_packets != old_camera_t_packets) // we have a new camera packet
    {
//      printf("confidence=%d, ", T_Packet_Data.confidence);
      // confidence still jumps to 84 periodically, but the rest of the data is good
      if (T_Packet_Data.confidence == 84) // Ignore this possibly bogus packet(somewhat fixed in camera.c)
      { // Print the suspect packet (84 = 'T')
        // every once in awhile we receive a T packet that has no data
        // It usually has a new data start flag (255) a few bytes after the "T" 
//        printf("mx=%d, my=%d, ", T_Packet_Data.mx, T_Packet_Data.my);
//        printf("x1=%d, y1=%d, ", T_Packet_Data.x1, T_Packet_Data.y1);
//        printf("x2=%d, y2=%d, ", T_Packet_Data.x2, T_Packet_Data.y2);
//        printf("pixels=%d\r", T_Packet_Data.pixels);
      }
      else
      {
        packet_err_cnt = 0;
	    if(T_Packet_Data.confidence >= CONF_THRESH)  // We are confident we see the target
	    {
//          printf("confidence=%d, ", T_Packet_Data.confidence);
//          printf("mx=%d, my=%d, ", T_Packet_Data.mx, T_Packet_Data.my);
//          printf("x1=%d, y1=%d, ", T_Packet_Data.x1, T_Packet_Data.y1);
//          printf("x2=%d, y2=%d, ", T_Packet_Data.x2, T_Packet_Data.y2);
//          printf("pixels=%d\r", T_Packet_Data.pixels);
//          printf("\r");

          search_init=TRUE; //Initialize the search in case we lose the target
		  error = (int)T_Packet_Data.mx - PAN_TARGET_PIXEL_DEFAULT;
		  //if(p3_sw_trig)
		  //	error *= -1;

          // camera update rate is ~10Hz so we are stuck with our motor choice for 4 motor update cycles before next update

          // Check if we're close enough for targeting
          // Targeting is given a larger deadband than tracking
		  if(abs(error) > TARGETING_DEADBAND)
	  	  {
			on_target = 0;
            // Indicate on the OI which direction the target is
            if (error > 0)
            {
              Relay2_green = ON;    // Target is to the left (or is it right? :)
              Relay2_red   = OFF;
              Last_Seen = LEFT;
            }
            else
            {
              Relay2_green = OFF;    // Target is to the right
              Relay2_red   = ON;
              Last_Seen = RIGHT;
            }
		  }
		  else  // we are close enough for shooting
		  {
			on_target = 1;
            // Indicate on the OI that we have target lock
            Relay2_green = ON;    /* Turn Relay2 green LED - OFF */
            Relay2_red   = ON;    /* Turn Relay2 red LED   - ON */
		  }
          Pan_By_Camera(pan_motor);          
          old_pan_motor = *pan_motor;
        }
        else // We do not see the target
        {
          on_target = FALSE;
          Relay2_green = OFF;    // Target is not visible
          Relay2_red   = OFF;
          // Attempt to search - simple back-n-forth
          Search_Pattern(pan_motor, search_init);
          old_pan_motor = *pan_motor;
          search_init=FALSE; // Only initialize if this is the first time searching
        }
      }
	}
    else   // No new camera packet - Check for the possibility of camera being offline
    {    // We MAY not be receiving camera packets
        if(packet_err_cnt >= PACKET_MISS_THRESH)
        {
            on_target = FALSE;  // It's been too long, override whatever the last valid packet claimed
        }
        else
        {
            packet_err_cnt++;
        }
    }
    //printf("Packet Count Smarter Spot, %d\r",camera_t_packets);
  }

    // Multiple on_target returns are necessary before we decide we have a target_lock
    if (old_camera_t_packets != camera_t_packets)
    {  // We have a fresh packet...
      if (on_target)
      { // and we're on target...
        if (target_hits > TARGET_HIT_THRESH) // a couple of hits in a row means we definitely have the target
           target_lock = TRUE;
        else
        {
           target_hits++;
           target_lock = FALSE;
        }
      }
      else
      {
        target_lock = FALSE;
        target_hits = 0;
      }
    }
      
    old_camera_t_packets = camera_t_packets;

  	return target_lock;
}

#if 0 // No tilt mechanism
char GimbalTrackTilt(void)
{
	signed int pix_err;
	signed int error;
	signed int cum_err;
	signed int last_pos;
	signed int d_err;
	signed int output;
	char init;
	char on_target;

	if(!init)
	{
		error = 0;
		cum_err = 0;
		last_pos = 0;
		output = 0;
		d_err = 0;
		on_target = 0;	
		init++;
	}
	else
	{
		pix_err = CAM_TILT_CENTER - PanTarget;
		error = CAM_TILT_POSITION - TiltPosTable[pix_err];

		if(abs(error) > CAM_DEADBAND)
		{
			cum_err += error;
			on_target = 0;
		}
		else
		{
			cum_err = output = 0;	
			on_target = 1;
		}
		d_err = last_pos - CAM_TILT_CENTER;
		output = (error * KP) + (cum_err * KI) - (d_err * KD);
	}

	return on_target;
}
#endif // No tilt mechanism

char GimbalTrack(void)
{
	char on_target = 0;
    int pan_motor;

	if(GimbalTrackPan(&pan_motor))
    {
//	if(GimbalTrackPan(&pan_motor) && GimbalTrackTilt())
		on_target = 1;
        PAN_MOTOR = pan_motor;
    }
	
	return on_target;
}

char Checklist(void)
{
	char items = 5;
	if(0 == 0)//BALL_SWITCH)
		items--;
	if(!PISTONUP)
		items--;
	if(0 == 0)//PRESSURE_GAUGE >= PRESSURE_REQ)
		items--;
	if(0 == 0)//DrumsAtSpeed())
		items--;
	if(PermToFire)
		items--;
    //printf("This shouldn't be running!\r");	
    //printf("Why not?\r");
	return items;
}

#if 0
void Semi_Auto(void)
{
	PanTarget  += ((ip3_x * 10) / 100); // Scale motors by 1/10th for offset
	TiltTarget += ((ip3_y * 10) / 100); // Scale motors...	

	if(GimbalTrack() && !Checklist()) // No errors, checklist returns 0, GimbalTrack() returns TRUE on lock.
		LAUNCH_RELAY = 1; // Deploy piston	
}	

void Auto_Fire(void)
{	
	PermToFire = 1;
	Semi_Auto();
}	

void Manual_Fire(void)
{
	if(!Checklist()) // No errors, checklist returns 0
		LAUNCH_RELAY = 1; // Deploy piston

	PAN_MOTOR  = ((ip3_x * 10) / 50); // Scale motors by 1/5th for precise manual targeting
	TILT_MOTOR = ((ip3_y * 10) / 50); // Scale motors...
}

#endif


/*******************************************************************************
* FUNCTION NAME: Pan_By_Camera
* PURPOSE:       Determine motor power to move pan to a position.
*                Treat one direction as separate and distinct from the opposite direction
*      pan_motor - suggested pan motor value to correct the error
*
*    See also in targeting.h all the PAN defines
*******************************************************************************/
#define KP_R 2/10    // Basic proportional adjustment (2/10 works)
#define KI_R 150/100  // 150/100 works
#define KD_R 5/10    // Was 5/10

#define KP_L 3/10    // Basic proportional adjustment (3/10 works)
#define KI_L 100/100  //  100/100 works
#define KD_L 0/10    // Was 0/10

#define CAM_DEADBAND 3 // When to stop panning was 4
#define IMAX 300  // 300 max = 127*KI

void Pan_By_Camera(int *pan_motor)
{
  int error, current_position, desired_position, temp;
  static int last_position=0, last_error=0, i=0, d=0;
  static char crossover_flag = FALSE;

  current_position = T_Packet_Data.mx;
  desired_position = PAN_TARGET_PIXEL_DEFAULT;
  if (current_position > 0) Last_Sighting=2; // Target is to the right
  if (current_position < 0) Last_Sighting=1; // Target is to the left

  // Proportional
  error = (current_position - desired_position);

  // Derivative
  d = last_error - error;
  last_error = error;

  // Integral
  if (error > 0 && crossover_flag)
  {
    i=0;
    crossover_flag = FALSE;
  }
  else if (error < 0 && !crossover_flag)
  {
    i=0;
    crossover_flag = TRUE;
  }
 
  if (current_position == last_position)
  { // We are not moving, so pick up the tempo just enough to get going
    if (i == 0) 
    {
        temp = PAN_MIN_SPEED_i_CMU; // Start i just below movement speed for quicker response
    }
    else
      temp = 1;

    if (current_position < desired_position)
    {
      i -= temp;
    }
    else
    {
      i += temp;
    }
    if (i >  IMAX) i =  IMAX;
    if (i < -IMAX) i = -IMAX;
  }
  last_position = current_position;

  if (((last_error < 0) && (error > 0)) || ((last_error > 0) && (error < 0)))
  { // we crossed over
    i = 0;
  }

  //------------------------

  if ((error < CAM_DEADBAND) && (error > -CAM_DEADBAND))  // Are we there yet?
  {
    *pan_motor = 0; // We are there - Stop, halt, pause, wait
    i = 0;
    last_error = 0;
  }
  else  // Not there yet...
  {
    // Test for commanded direction
    if (error > 0)
    {
      printf("PanPot=,%d,KP=,%d,KI=,%d,KD=,%d,\r", PANPOT, (error * KP_L),(i * KI_L),(d * -KD_L));
      *pan_motor = (error * KP_L) + (i * KI_L) - (d * KD_L);
      if (*pan_motor > PAN_TOP_SPEED_LEFT) *pan_motor = PAN_TOP_SPEED_LEFT;
    }
    else
    {
      printf("PanPot=,%d,KP=,%d,KI=,%d,KD=,%d,\r", PANPOT,(error * KP_R),(i * KI_R),(d * -KD_R));
      *pan_motor = (error * KP_R) + (i * KI_R) - (d * KD_R);
      if (*pan_motor < -PAN_TOP_SPEED_RIGHT) *pan_motor = -PAN_TOP_SPEED_RIGHT;
    }

    *pan_motor *= PAN_MOTOR_DIRECTION;
//printf(" Cam_Pan limit check \r");
    // Know our limits...
    if ((PANPOT <= PANPOTMIN) && (*pan_motor > 0))
    {
      *pan_motor = 0;
    }
    else if ((PANPOT >= PANPOTMAX) && (*pan_motor < 0))
    {
      *pan_motor = 0;
    }
  }
//  printf("desired_position=%d, current_position=%d, pan_motor=%d, i=%d, d=%d\r", desired_position, current_position,*pan_motor, i, d);
}


/*******************************************************************************
* FUNCTION NAME: Search_Pattern
* PURPOSE:       Move the gimbal back and forth while the target is not visible.
*
*******************************************************************************/
void Search_Pattern(int *pan_motor, char init)
{
  static char direction=LEFT; // Start by heading right towards the "long" side
  static char full_search=FALSE;
  signed long rawTargetHeading;
  long error;

  if (init)
  { // Reset for a fresh search
    direction = Last_Seen; // Where was the target last detected?
    full_search = FALSE;
  }

#if 0 //Untested
  if (full_search)
  { 
   // We've searched all the territory we possibly can. Reorient the robot based on the gyro.
    rawTargetHeading = GYRO_DEG_RAW(90);  // Turn towards our target light
    error = RawHeading - rawTargetHeading;
    if ((error > GTURN_DEADBAND_BIG) || (error < -GTURN_DEADBAND_BIG)))
    { // Keep turning
      WHL_R1 = GyroTurnRaw(rawTargetHeading);
      WHL_L1 = -WHL_R1;
    }
    else 
    {
      // Reset for a fresh search
      direction = RIGHT;
      full_search = FALSE;
      WHL_L1 = WHL_R1 = 0;
    }
  }
#endif

  if (direction == RIGHT)
  {
    if (PANPOT >= (PANPOTMAX-80))
    {
      direction = LEFT; // Reverse direction
      *pan_motor = 0; // Stop motor for one cycle (abrupt but not quite so abrupt)
                      // May want to stretch this delay in reversing our engines
    }
    else // Keep going RIGHT
    {
//      *pan_motor = -PAN_TOP_SPEED_RIGHT;
      *pan_motor = -PAN_SEARCH_SPEED;
    }
  }
  else // Direction is 1 or to the left
  {
    if (PANPOT <= PANPOTMIN+80)
    {
      direction = RIGHT; // Reverse direction
      *pan_motor = 0; // Stop motor for one cycle (abrupt but not quite so abrupt)
                      // May want to stretch this delay in reversing our engines
      full_search = TRUE; // We've searched all the territory we possibly can. Time to rotate the robot.

    }
    else  // Keep going LEFT
    {
//      *pan_motor = PAN_TOP_SPEED_LEFT;
      *pan_motor = PAN_SEARCH_SPEED;
    }
  }
}
