/*******************************************************************************
* FILE NAME: user_routines_fast.c
*
* DESCRIPTION:
*  Interrupt handlers and fast loop updaters based on interrupts are processed here.
*
*******************************************************************************/


#include <stdio.h>
#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "user_routines.h"
#include "serial_ports.h"
#include "clock_routines.h"
#include "gyro.h"

unsigned int counter = 0;
unsigned long number_o_loops=0;

/*******************************************************************************
* FUNCTION NAME: InterruptVectorLow
* PURPOSE:       Low priority interrupt vector
* DO NOT MODIFY OR DELETE THIS FUNCTION 
*******************************************************************************/
#pragma code InterruptVectorLow = LOW_INT_VECTOR
void InterruptVectorLow (void)
{
  _asm
    goto InterruptHandlerLow  /*jump to interrupt routine*/
  _endasm
}


/*******************************************************************************
* FUNCTION NAME: InterruptHandlerLow
* PURPOSE:       Low priority interrupt handler
* Users are only allowed external low priority interrupts 
*******************************************************************************/
#pragma code
#pragma interruptlow InterruptHandlerLow save=PROD,section("MATH_DATA"),section(".tmpdata")

#define RISING_EDGE 0
#define FALLING_EDGE 1

void InterruptHandlerLow ()     
{
  unsigned char Port_B;
  unsigned char Port_B_Delta;       
  static unsigned char Old_Port_B = 0xFF; // state of port b the last time
											// this function was called
  static char LeftEncoderEdge=0, RightEncoderEdge=0;

  // These are the serial communications interrupts developed by Kevin Watson
	if (PIR1bits.RC1IF && PIE1bits.RC1IE) // rx1 interrupt?
	{
		#ifdef ENABLE_SERIAL_PORT_ONE_RX
		Rx_1_Int_Handler(); // call the rx1 interrupt handler (in serial_ports.c)
		#endif
	}                              
	else if (PIR3bits.RC2IF && PIE3bits.RC2IE) // rx2 interrupt?
	{
		#ifdef ENABLE_SERIAL_PORT_TWO_RX
		Rx_2_Int_Handler(); // call the rx2 interrupt handler (in serial_ports.c)
		#endif
	} 
	else if (PIR1bits.TX1IF && PIE1bits.TX1IE) // tx1 interrupt?
	{
		#ifdef ENABLE_SERIAL_PORT_ONE_TX
		Tx_1_Int_Handler(); // call the tx1 interrupt handler (in serial_ports.c)
		#endif
	}                              
	else if (PIR3bits.TX2IF && PIE3bits.TX2IE) // tx2 interrupt?
	{
		#ifdef ENABLE_SERIAL_PORT_TWO_TX
		Tx_2_Int_Handler(); // call the tx2 interrupt handler (in serial_ports.c)
		#endif
	}


  /*** Service Internal timer interrupts ***/
 
  if (PIR2bits.TMR3IF)	/* TIMER 3 INTERRUPT */
  {
    /** This provides us with a clock for timing events **/

    PIR2bits.TMR3IF = 0;	/* Clear Timer interrupt flag */
    WriteTimer3(CLOCKOFFSET);  /* Reset Timer to overflow in 4ms */
    Clockms += 4; /* milliseconds */
    Clockds = Clockms / 100;
    Clocksec = Clockms / 1000;
    GyroTicks++;
  }


  /*** Service External Interrupts ***/

  if (INTCON3bits.INT2IF) // external interrupt 1
  { 
	INTCON3bits.INT2IF = 0; // clear the interrupt flag
    // Could double encoder resolution by also putting "B" on an interrupt
    // Could double again by interrupting on falling edge too
    // One encoder revolution (128 interrupts) = ?" of travel
    // One encoder "tick" = .23"
    // 2006 - 85.3 ticks per wheel revolution 19 5/8" on traction wheel 
    // 19 1/8" on slick wheels

    // Need a state machine too to prevent flicker
	if(LEFT_B_ENCODER)  //Quad encoder B signal
    {
		Left_Encoder--;
        Total_Left_Encoder_Count--;
    }
	else
    {
		Left_Encoder++;
        Total_Left_Encoder_Count++;
    }

    /* Make sure the other encoder hasn't also fired */
    if (INTCON3bits.INT3IF == 1)
    {
        INTCON3bits.INT3IF = 0; // clear the interrupt flag
 		if(RIGHT_B_ENCODER)  //Quad encoder B signal
        {
		  	Right_Encoder++;
		    Total_Right_Encoder_Count++;
        }
		else
        {
		  	Right_Encoder--;
			Total_Right_Encoder_Count--;
        }
     }
  }
  else if (INTCON3bits.INT3IF) // external interrupt 2 
  {
	INTCON3bits.INT3IF = 0; // clear the interrupt flag
	// if used, call the interrupt 2 handler here
	if(RIGHT_B_ENCODER)  //Quad encoder B signal
    {
		Right_Encoder++;
		Total_Right_Encoder_Count++;
    }
	else
    {
		Right_Encoder--;
		Total_Right_Encoder_Count--;
    }

    /* Make sure the other encoder hasn't also fired */
    if (INTCON3bits.INT2IF == 1)
    {
	    INTCON3bits.INT2IF = 0; // clear the interrupt flag
		if(LEFT_B_ENCODER)  //Quad encoder B signal
        {
			Left_Encoder--;
            Total_Left_Encoder_Count--;
        }
	    else
        {
			Left_Encoder++;
            Total_Left_Encoder_Count++;
        }
     }
  }
  else if (INTCONbits.RBIF) // external interrupts 3 through 6
  {
	Port_B = PORTB; // remove the "mismatch condition" by reading port b            
	INTCONbits.RBIF = 0; // clear the interrupt flag
	Port_B_Delta = Port_B ^ Old_Port_B; // determine which bits have changed
	Old_Port_B = Port_B; // save a copy of port b for next time around
	 
	if(Port_B_Delta & 0x10) // Check interrupt 3
	{
	}
	if(Port_B_Delta & 0x20) // Check interrupt 4
	{
	}
	if(Port_B_Delta & 0x40) // Check interrupt 5
	{
	}
	if(Port_B_Delta & 0x80) // Check interrupt 6
	{
	}
  }
}



/*******************************************************************************
* FUNCTION NAME: Process_Data_From_Local_IO
* PURPOSE:       Execute user's realtime code.
* You should modify this routine by adding code which you wish to run fast.
* It will be executed every program loop, and not wait for fresh data 
* from the Operator Interface.
* CALLED FROM:   main.c
*******************************************************************************/
void Process_Data_From_Local_IO(void)
{
  /* Add code here that you want to be executed every program loop. */
  Gyro();
//  counter = (counter + ReadTimer1()) / 2;
//  number_o_loops++;
}


/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
